@@ -3,7 +3,7 @@
 //! This module contains specific libc Linux system calls to be performed when
 //! calling into [crate::platforms] on a Linux OS.
 
-use const_format::concatcp;
+use cfg_if::cfg_if;
 use cstr::cstr;
 use nix::{
     errno::Errno,
@@ -12,6 +12,11 @@ use nix::{
         open,
     },
     ioctl_read,
+    libc::{
+        STDERR_FILENO,
+        STDIN_FILENO,
+        STDOUT_FILENO,
+    },
     sys::{
         reboot::{
             RebootMode,
@@ -67,14 +72,14 @@ use crate::{
 
 
 /// Used as the 0th argument to serviced in an `execv` call.
-const SERVICED_ARG0: &'static str = "serviced";
+const SERVICED_ARG0: &'static CStr = cstr!("serviced");
 
 /// Used as the 1st argument to serviced in an `execv` call.
 ///
 /// All initd-compatible programs should use this same constant value to indicate
 /// their communication specification. serviced can parse the sent value and
 /// allow communication accordingly.
-#[allow(dead_code)] const SERVICED_ARG1: &'static str = "initd_v0";
+#[allow(dead_code)] const SERVICED_ARG1: &'static CStr = cstr!("initd_v0");
 
 /// Path where serviced is located. Used in the `execv` call to actually execute
 /// serviced.
@@ -83,7 +88,7 @@ const SERVICED_ARG0: &'static str = "serviced";
 /// actually puts the executable in `/sbin/serviced`. Otherwise, you must maintain a
 /// patch changing `SERVICED_PATH` to the appropriate path (e.g. `/serviced`,
 /// `/bin/serviced`, or `/usr/bin/serviced`).
-const SERVICED_PATH: &'static str = concatcp!("/sbin/", SERVICED_ARG0);
+const SERVICED_PATH: &'static CStr = cstr!("/sbin/serviced");
 
 /// Error message used in case `SERVICE_PATH` is not able to be executed by `execv`.
 /// This can be caused by not having serviced installed (although distributions should
@@ -91,7 +96,7 @@ const SERVICED_PATH: &'static str = concatcp!("/sbin/", SERVICED_ARG0);
 /// not having it installed with the proper executable permissions (should never
 /// happen unless serviced is improperly packaged by the distribution), or a more
 /// obscure error.
-const SERVICED_ERROR: &'static str = concatcp!("unable to execute serviced in ", SERVICED_PATH);
+const SERVICED_ERROR: &'static str = "unable to execute serviced";
 
 /// In how many seconds the alarm should be scheduled.
 ///
@@ -139,28 +144,28 @@ ioctl_read! {
 ///
 /// The following sanity checks should be made in Linux:
 /// * We should be PID 1. Otherwise, an error should be returned.
-///     * In debug mode, we can just emit a warning and continue.
-pub(crate) fn initial_sanity_check() -> Result<OpaqueSanityCheckResult, PrintableErrno> {
+///     * With `debug-notpid1`, we can just emit a warning and continue.
+pub(crate) fn initial_sanity_check() -> Result<OpaqueSanityCheckResult, PrintableErrno<String>> {
     let pid = getpid().as_raw();
-    Ok(
-        if pid != 1 {
-            // If we aren't PID 1, assume stdin/stdout/stderr to all be open, so no problem in
-            // using precisej-printable-errno here.
-            let e = printable_error(PROGRAM_NAME, format!("expected PID 1, got {}.", pid));
+    if pid != 1 {
+        // If we aren't PID 1, assume stdin/stdout/stderr to all be open, so no problem in
+        // using precisej-printable-errno here.
+        let e = printable_error(PROGRAM_NAME, format!("expected PID 1, got {}.", pid));
 
-            // Running initd as a regular process shouldn't be permitted in release mode, but
-            // it makes for easy testing in debug mode.
-            if cfg!(debug_assertions) {
+        // Running initd as a regular process shouldn't be permitted in release mode, so it's
+        // hidden behind the 'debug-notpid1' feature flag.
+        cfg_if! {
+            if #[cfg(feature = "debug-notpid1")] {
                 e.eprint();
                 printable_error(PROGRAM_NAME, "continuing anyway...".to_string()).eprint();
-                OpaqueSanityCheckResult(false)
+                Ok(OpaqueSanityCheckResult(false))
             } else {
-                return Err(e)
+                Err(e)
             }
-        } else {
-            OpaqueSanityCheckResult(true)
         }
-    )
+    } else {
+        Ok(OpaqueSanityCheckResult(true))
+    }
 }
 
 /// Initial setup for Linux.
@@ -171,45 +176,45 @@ pub(crate) fn initial_sanity_check() -> Result<OpaqueSanityCheckResult, Printabl
 ///     * Set read_fd as stdin.
 ///     * Set write_fd as stdout and stderr.
 /// * Change directory to `/`.
-pub(crate) fn initial_setup(results: &OpaqueSanityCheckResult) -> Result<(), PrintableErrno> {
+pub(crate) fn initial_setup(results: &OpaqueSanityCheckResult) -> Result<(), PrintableErrno<&'static str>> {
     let is_pid1 = results.0;
     if is_pid1 {
         // Since neither stdin/stdout/stderr is assumed to be open, we can't use
         // precisej-printable-errno just yet.
-        // In other functions with errno, runit simply repeats on failure after 5 seconds,
-        // so we will do something similar here.
-        const STDIN: RawFd = 0;
-        const STDOUT: RawFd = 1;
-        const STDERR: RawFd = 2;
+        // runit repeats syscalls on failure after 5 seconds, so we will do something
+        // similar here.
 
-        //noinspection DuplicatedCode
-        let read_fd = loop {
-            match open("/dev/console", OFlag::O_RDONLY, Mode::empty()) {
-                Ok(fd) => break fd,
-                Err(_) => {
-                    sleep(5);
-                }
-            }
-        };
-        while dup2(read_fd, STDIN).is_err() { sleep(5); }
-        if read_fd != STDIN {
+        let read_fd = open_console(OFlag::O_RDONLY);
+        dup2_retry(read_fd, STDIN_FILENO);
+        if read_fd != STDIN_FILENO {
             close(read_fd).ok(); // ignore error
         }
 
-        //noinspection DuplicatedCode
-        let write_fd = loop {
-            match open("/dev/console", OFlag::O_WRONLY, Mode::empty()) {
-                Ok(fd) => break fd,
-                Err(_) => {
-                    sleep(5);
-                }
-            }
-        };
-        while dup2(write_fd, STDOUT).is_err() { sleep(5); }
-        while dup2(write_fd, STDERR).is_err() { sleep(5) ;}
-        if write_fd != STDOUT && write_fd != STDERR {
+        let write_fd = open_console(OFlag::O_WRONLY);
+        dup2_retry(write_fd, STDOUT_FILENO);
+        dup2_retry(write_fd, STDERR_FILENO);
+        if write_fd != STDOUT_FILENO && write_fd != STDERR_FILENO {
             close(write_fd).ok(); // ignore error
         }
+
+        #[inline]
+        fn open_console(oflag: OFlag) -> RawFd {
+            loop {
+                match open("/dev/console", oflag, Mode::empty()) {
+                    Ok(fd) => break fd,
+                    Err(_) => sleep(5),
+                };
+            }
+        }
+        #[inline]
+        fn dup2_retry(oldfd: RawFd, newfd: RawFd) {
+            loop {
+                match dup2(oldfd, newfd) {
+                    Ok(_) => break,
+                    Err(_) => sleep(5),
+                };
+            }
+        }
     }
 
     // At this point precisej-printable-errno should work as stdin/stdout/stderr should
@@ -248,7 +253,7 @@ pub(crate) fn alarm_clear() {
 /// ourselves. Therefore `sync(2)` must be called as a precautionary measure.
 ///
 /// After calling `sync(2)`, we call `reboot(2)` with `RB_POWER_OFF`.
-pub(crate) fn power_off() -> Result<Infallible, PrintableErrno> {
+pub(crate) fn power_off() -> Result<Infallible, PrintableErrno<&'static str>> {
     sync();
     reboot(RebootMode::RB_POWER_OFF).printable(PROGRAM_NAME, "unable to shutdown")
 }
@@ -261,7 +266,7 @@ pub(crate) fn power_off() -> Result<Infallible, PrintableErrno> {
 /// ourselves. Therefore `sync(2)` must be called as a precautionary measure.
 ///
 /// After calling `sync(2)`, we call `reboot(2)` with `RB_AUTOBOOT`.
-pub(crate) fn reboot_autoboot() -> Result<Infallible, PrintableErrno> {
+pub(crate) fn reboot_autoboot() -> Result<Infallible, PrintableErrno<&'static str>> {
     sync();
     reboot(RebootMode::RB_AUTOBOOT).printable(PROGRAM_NAME, "unable to reboot")
 }
@@ -294,7 +299,7 @@ impl OpaqueSigSet {
     ///
     /// **NOTE**: even though our set contains all signals, `SIGKILL` and `SIGSTOP`
     /// can't be blocked and they are silently ignored in `sigprocmask(2)`.
-    pub(crate) fn block(&self) -> Result<(), PrintableErrno> {
+    pub(crate) fn block(&self) -> Result<(), PrintableErrno<&'static str>> {
         sigprocmask(SIG_BLOCK, Some(&self.set), None)
             .printable(PROGRAM_NAME, "unable to block signals")
     }
@@ -303,7 +308,7 @@ impl OpaqueSigSet {
     ///
     /// **NOTE**: even though our set contains all signals, `SIGKILL` and `SIGSTOP`
     /// can't be blocked and they are silently ignored in `sigprocmask(2)`.
-    pub(crate) fn unblock(&self) -> Result<(), PrintableErrno> {
+    pub(crate) fn unblock(&self) -> Result<(), PrintableErrno<&'static str>> {
         sigprocmask(SIG_UNBLOCK, Some(&self.set), None)
             .printable(PROGRAM_NAME, "unable to unblock signals in child")
     }
@@ -313,7 +318,7 @@ impl OpaqueSigSet {
     ///
     /// **NOTE**: even though our set contains all signals, `SIGKILL` and `SIGSTOP`
     /// can't be blocked and they are silently ignored in `sigprocmask(2)`.
-    pub(crate) fn wait_for_next(&self) -> Result<Option<platforms::ProcSignal>, PrintableErrno> {
+    pub(crate) fn wait_for_next(&self) -> Result<Option<platforms::ProcSignal>, PrintableErrno<&'static str>> {
         match self.set.wait().printable(PROGRAM_NAME, "unable to await new process signals")? {
             Signal::SIGUSR1 => Ok(Some(platforms::ProcSignal::PowerOff)),
             Signal::SIGCHLD => Ok(Some(platforms::ProcSignal::ReapChild)),
@@ -355,7 +360,7 @@ impl OpaqueServicedHandle {
     ///     * In the parent:
     ///         * Close serviced's end of the pipe, as initd doesn't need it.
     ///         * Return a communication handle to the recently spawned serviced.
-    pub(crate) fn spawn_serviced(set: &OpaqueSigSet, setup: OpaqueSanityCheckResult) -> Result<OpaqueServicedHandle, PrintableErrno> {
+    pub(crate) fn spawn_serviced(set: &OpaqueSigSet, setup: OpaqueSanityCheckResult) -> Result<OpaqueServicedHandle, PrintableErrno<&'static str>> {
         let (read_exit_pipe, write_exit_pipe) = pipe2(OFlag::O_NONBLOCK)
             .printable(PROGRAM_NAME, "unable to create critical communication channel with serviced")?;
 
@@ -387,33 +392,27 @@ impl OpaqueServicedHandle {
                     .printable(PROGRAM_NAME, "serviced: unable to close duplicate communication channel end with initd")
                     .ok_or_eprint_signal_safe();
 
-                // All we have to do now is change session / process group id, open /dev/console/,
+                // All we have to do now is change session / process group id, open /dev/console,
                 // and hand over execution to serviced. Realistically the only error that could
                 // arise is SERVICED_PATH not being found (root not properly mounted or no serviced/
                 // serviced-compatible program installed), so just begrudgingly bail and hope any
-                // distro maintainer fixes whatever crap package manager dependency management they
-                // have that permits installing ONLY initd.
+                // distro maintainer fixes whatever crap caused this.
                 setsid()
                     .printable(PROGRAM_NAME, "unable to change session id")
                     .ok_or_eprint_signal_safe();
 
                 if is_pid1 {
-                    const STDIN: RawFd = 0;
-                    const STDERR: RawFd = 2;
-
                     if let Ok(fd) = open("/dev/console", OFlag::O_RDWR, Mode::empty()) {
                         tiocsctty(fd, std::ptr::null_mut()).ok(); // ignore error
 
-                        write(1, b"hello child 2\n").ok();
-
-                        while dup2(fd, STDIN).is_err() { sleep(5); }
-                        if fd > STDERR {
+                        while dup2(fd, STDIN_FILENO).is_err() { sleep(5); }
+                        if fd > STDERR_FILENO {
                             close(fd).ok(); // ignore error
                         }
                     };
                 }
 
-                let err = execv(cstr!(SERVICED_PATH), &[cstr!(SERVICED_ARG0), cstr!(SERVICED_ARG1), serviced_exit_pipe_arg2])
+                let err = execv(SERVICED_PATH, &[SERVICED_ARG0, SERVICED_ARG1, serviced_exit_pipe_arg2])
                     .printable(PROGRAM_NAME, SERVICED_ERROR)
                     .bail(7);
 
@@ -423,7 +422,9 @@ impl OpaqueServicedHandle {
                     .printable(PROGRAM_NAME, "serviced: unable to clean up communication channel end with initd")
                     .ok_or_eprint_signal_safe();
                 err.unwrap_or_eprint_signal_safe_exit();
-                unreachable!()
+
+                // unreachable because we already exited
+                std::hint::unreachable_unchecked()
             }
             Err(errno) => {
                 // fork() failing should be a fatal error, so bubble up a printable error.
@@ -439,11 +440,11 @@ impl OpaqueServicedHandle {
 
     /// Await the next child process. The result is of type [platforms::WaitStatus] to
     /// distinguish what should be done.
-    pub(crate) fn wait_next_child<S: ServicedPid>(serviced: &S) -> Result<platforms::WaitStatus, PrintableErrno> {
+    pub(crate) fn wait_next_child<S: ServicedPid>(serviced: &S) -> Result<platforms::WaitStatus, PrintableErrno<String>> {
         let status = match waitpid(None, Some(WaitPidFlag::WNOHANG)) {
             Ok(status) => status,
             Err(Errno::ECHILD) => return Ok(platforms::WaitStatus::BreakDefault), // WNOHANG: nothing else to wait for
-            Err(e) => return Err(e).printable(PROGRAM_NAME, "unable to wait for child")
+            Err(e) => return Err(e).printable(PROGRAM_NAME, "unable to wait for child".to_string())
         };
         let serviced_pid = serviced.pid();
         match status {
@@ -478,7 +479,7 @@ impl OpaqueServicedHandle {
     }
 
     /// Send a message through the handle to serviced to cleanly exit, and then close the handle.
-    pub(crate) fn send_exit_message(&mut self) -> Result<OpaqueClosingServicedInstance, PrintableErrno> {
+    pub(crate) fn send_exit_message(&mut self) -> Result<OpaqueClosingServicedInstance, PrintableErrno<&'static str>> {
         write(self.exit_pipe, &[1]).printable(PROGRAM_NAME, "unable to send exit message to serviced")?;
         close(self.exit_pipe).printable(PROGRAM_NAME, "unable to close message pipe with serviced")?;
         self.exit_pipe = -1;
